package be.iminds.ilabt.jfed.experimenter_gui.slice.espec.model;

import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.util.ESpecLogListener;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLog;
import be.iminds.ilabt.jfed.espec.util.UploadProgressTracker;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import javafx.application.Platform;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;

/**
 * ESpecLogModel
 */
public class ESpecLogModel implements ESpecLogListener {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ESpecLogModel.class);

    public final static String AQUIRE_RESOURCES_STEP_ID = "Acquire Resources";
    public final static String SETUP_SOFTWARE_STEP_ID = ExperimentState.SETUP_SOFTWARE.name();

    private final static class ESpecStateId {
        @Nonnull private final String baseId;
        @Nullable private final String subId;

        public ESpecStateId(@Nullable ExperimentState experimentState, @Nullable String especStepId) {
            this.baseId = findBaseId(experimentState, especStepId);
            this.subId = findSubId(experimentState, especStepId);
            LOG.debug("Mapped IDs: "+experimentState+" "+especStepId+"  -->  "+baseId+" "+subId);
            assert !(Objects.equals(baseId, "ALLOCATING") && subId == null);
        }
        private ESpecStateId(@Nonnull String baseId, @Nullable String subId) {
            assert !(Objects.equals(baseId, "ALLOCATING") && subId == null);
            this.baseId = baseId;
            this.subId = subId;
        }

        public ESpecStateId(@Nullable ExperimentState experimentState) {
            this(experimentState, null);
        }

        private static boolean partOfResourceSetup(@Nonnull ExperimentState experimentState) {
            return experimentState == ExperimentState.ALLOCATING ||
                    experimentState == ExperimentState.PROVISIONING ||
                    experimentState == ExperimentState.WAIT_FOR_READY ||
                    experimentState == ExperimentState.TESTING_CONNECTIVITY;
        }

        @Nonnull
        private String findBaseId(@Nullable ExperimentState experimentState, @Nullable String especStepId) {
            if (experimentState == null) {
                if (especStepId != null)
                    return ExperimentState.SETUP_SOFTWARE.name();
                //noinspection ConstantConditions
                throw new RuntimeException("Unknown state: " +
                        "experimentState=" + experimentState + " " +
                        "especStepId=" + especStepId);
            }
            if (partOfResourceSetup(experimentState))
                return AQUIRE_RESOURCES_STEP_ID;
            return experimentState.name();
        }
        @Nullable
        public String findSubId(@Nullable ExperimentState experimentState, @Nullable String especStepId) {
            if (experimentState == null) {
                if (especStepId != null)
                    return especStepId;
                //noinspection ConstantConditions
                throw new RuntimeException("Unknown state: " +
                        "experimentState=" + experimentState + " " +
                        "especStepId=" + especStepId);
            }
            if (partOfResourceSetup(experimentState))
                return experimentState.name();

            return especStepId;
        }

        @Nonnull public String getBaseId() {
            return baseId;
        }
        @Nullable public String getSubId() {
            return subId;
        }
        public boolean isChild() {
            return subId != null;
        }
        @Nonnull public ESpecStateId getParentId() {
            return new ESpecStateId(baseId, null);
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;
            ESpecStateId that = (ESpecStateId) o;
            return Objects.equals(baseId, that.baseId) &&
                    Objects.equals(subId, that.subId);
        }

        @Override
        public int hashCode() {
            return Objects.hash(baseId, subId);
        }

        @Override
        public String toString() {
            return "ESpecStateId{" + baseId + ", " + subId + '}';
        }
    }

    private final Map<ESpecStateId, ESpecStateItem> internalESpecStateItems = new ConcurrentHashMap<>();

    private final Map<Long, ESpecLogItem> eSpecLogItemsById;
    private final ObservableList<ESpecLogItem> eSpecLogItems;
    private final ObservableList<ESpecStateItem> eSpecStateItems;

    private final ObjectProperty<ESpecStateItem> selectedESpecStateItem;

    @Nonnull private final ESpecLogger logger;
    @Nonnull private final Experiment experiment;

    public ESpecLogModel(@Nonnull Experiment experiment) {
        this.logger = experiment.getExperimentSpecificationLogger();
        this.experiment = experiment;

        this.eSpecLogItemsById = new TreeMap<>();
        this.eSpecLogItems = FXCollections.observableArrayList();
        this.eSpecStateItems = FXCollections.observableArrayList();
        this.selectedESpecStateItem = new SimpleObjectProperty<>();

        experiment.addESpecLogListener(this);
        experiment.experimentStateProperty().addListener(
                (observableValue, oldVal, newVal) -> {
                    setPreviousRunningToSuccess(newVal, null);
                    boolean success = !isFailingState(newVal);
                    ESpecStateItem.ESpecStateStatus targetState = success ? (
                            newVal.isActiveState() && newVal != ExperimentState.READY ?
                                    ESpecStateItem.ESpecStateStatus.RUNNING :
                                    ESpecStateItem.ESpecStateStatus.SUCCESS) :
                            ESpecStateItem.ESpecStateStatus.FAILURE;
                    ESpecStateItem cur = getState(newVal, null, targetState);
                    cur.setStatus(targetState);
                    addLog()
                            .setText("Reached State " + newVal.name())
                            .setFinalStatus(success)
                            .setESpecStateItem(cur)
                            .addOrUpdate();

                    LOG.debug("Handled state change to "+newVal.name());
                    dumpDebug();
                }
        );

        addInitialStates();
    }

    private void addInitialStates() {
        getState(ExperimentState.ALLOCATING, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.PROVISIONING, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.WAIT_FOR_READY, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.TESTING_CONNECTIVITY, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.SETUP_SOFTWARE, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.SETUP_SOFTWARE, "Dir", ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.SETUP_SOFTWARE, "Upload", ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.SETUP_SOFTWARE, "Execute", ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.SETUP_SOFTWARE, "Ansible", ESpecStateItem.ESpecStateStatus.FUTURE);
        getState(ExperimentState.READY, null, ESpecStateItem.ESpecStateStatus.FUTURE);
        dumpDebug();
    }

    private void dumpDebug() {
        LOG.debug("All states:");
        Set<Map.Entry<ESpecStateId, ESpecStateItem>> all = internalESpecStateItems.entrySet();
        for (Map.Entry<ESpecStateId, ESpecStateItem> e : all) {
            ESpecStateId id = e.getKey();
            ESpecStateItem item = e.getValue();
            LOG.debug("State: "+id+" -> "+item.getId()+"    children="+
                    item.getChildren().stream()
                            .map(ESpecStateItem::getId)
                            .collect(Collectors.joining(",")));
        }
    }

    /**
     * Creates the state if not existing. Thread safe.
     * JavaFX thread will automatically be notified later when needed.
     *
     * ESpecStateItem of parent and child will be created when needed.
     * */
    @Nonnull
    private ESpecStateItem getState(@Nullable ExperimentState experimentState,
                                    @Nullable String especStepId,
                                    @Nonnull ESpecStateItem.ESpecStateStatus initialStatus) {
        ESpecStateId id = new ESpecStateId(experimentState, especStepId);
        return getState(id, initialStatus);
    }
    @Nonnull
    private ESpecStateItem getState(@Nonnull ESpecStateId id,
                                    @Nonnull ESpecStateItem.ESpecStateStatus initialStatus) {
        assert !(Objects.equals(id.getBaseId(), "ALLOCATING") && id.getSubId() == null);

        LOG.debug("getState("+id+", "+initialStatus+")");
        if (id.isChild()) {
            ESpecStateId parentId = id.getParentId();
            LOG.debug("getState() -> handle child with parentId="+parentId);
            ESpecStateItem parent = getState(parentId, initialStatus);
            LOG.debug("getState() -> parent.getId()="+parent.getId()+" parent.getChildren="+
                    parent.getChildren().stream()
                            .map(ESpecStateItem::getId)
                            .collect(Collectors.joining(",")));

            AtomicBoolean wasAdded = new AtomicBoolean(false);
            ESpecStateItem res = internalESpecStateItems.computeIfAbsent(id, (key) -> {
                ESpecStateItem toAdd = new ESpecStateItem(id.subId, initialStatus);
                wasAdded.set(true);
                return toAdd;
            });
            if (wasAdded.get()) {
                LOG.debug("getState() -> Adding child "+res.getId()+" ("+id+")"+" to "+parent.getId()+" ("+parentId+")");
                Platform.runLater(() -> {
                    parent.getChildren().add(res);
                });
            } else {
                LOG.debug("getState() -> child existed");
            }
            return res;
        } else {
            LOG.debug("getState() -> handle parent");
            AtomicBoolean wasAdded = new AtomicBoolean(false);
            ESpecStateItem res = internalESpecStateItems.computeIfAbsent(id, (key) -> {
                ESpecStateItem toAdd = new ESpecStateItem(id.baseId, initialStatus);
                wasAdded.set(true);
                return toAdd;
            });
            if (wasAdded.get()) {
                LOG.debug("getState() -> Adding parent "+res.getId()+" ("+id+")");
                Platform.runLater(() -> {
                    eSpecStateItems.add(res);
                });
            } else {
                LOG.debug("getState() -> parent existed");
            }
            return res;
        }
    }
    private void setPreviousRunningToSuccess(ExperimentState experimentState, String especStepId) {
        setPreviousRunningToSuccess(new ESpecStateId(experimentState, especStepId));
    }
    private void setPreviousRunningToSuccess(ESpecStateId id) {
        internalESpecStateItems.values()
                .stream()
                .filter(esi -> esi.getId() != id.baseId)
                .filter(esi -> id.subId == null || esi.getId() != id.subId)
                .filter(esi -> esi.getStatus() == ESpecStateItem.ESpecStateStatus.RUNNING)
                .forEach(esi -> esi.setStatus(ESpecStateItem.ESpecStateStatus.SUCCESS));
    }

    private void setStatus(ExperimentState experimentState, String especStepId, ESpecStateItem.ESpecStateStatus newStatus) {
        ESpecStateId id = new ESpecStateId(experimentState, especStepId);
        setStatus(id, newStatus);
    }
    private void setStatus(ESpecStateId id, ESpecStateItem.ESpecStateStatus newStatus) {
        ESpecStateItem item = getState(id, newStatus);
        item.setStatus(newStatus);
    }


    public ObservableList<ESpecLogItem> getESpecLogItems() {
        return eSpecLogItems;
    }

    public ObservableList<ESpecStateItem> getESpecStateItems() {
        return eSpecStateItems;
    }

    @Nullable
    public ESpecStateItem getSelectedESpecStateItem() {
        return selectedESpecStateItem.get();
    }

    public ObjectProperty<ESpecStateItem> selectedESpecStateItemProperty() {
        return selectedESpecStateItem;
    }

    public void setSelectedESpecStateItem(@Nullable ESpecStateItem selectedESpecStateItem) {
        this.selectedESpecStateItem.set(selectedESpecStateItem);
    }
    public void clearSelectedESpecStateItem() {
        this.selectedESpecStateItem.set(null);
    }

    public static boolean isFailingState(@Nonnull ExperimentState state) {
        return state == ExperimentState.FAILING ||
                state == ExperimentState.FAILED ||
                state == ExperimentState.UNKNOWN || //considered failing here
                state == ExperimentState.TIMEOUT_WAITING;
    }
    public static boolean isFailingState(@Nonnull String stateName) {
        return stateName.equalsIgnoreCase(ExperimentState.FAILING.name()) ||
                stateName.equalsIgnoreCase(ExperimentState.FAILED.name()) ||
                stateName.equalsIgnoreCase(ExperimentState.UNKNOWN.name()) || //considered failing here
                stateName.equalsIgnoreCase(ExperimentState.TIMEOUT_WAITING.name());
    }

    private class ESpecLogItemAdder {
        private final long id;

        private String text;
        private TaskStatusIndicator.Status status;

        private ESpecStep especStep;
        private LimitedLiveLog limitedLiveLog;
        private UploadProgressTracker uploadProgressTracker;

        //info on ESpecStateItem to log to
        private ESpecStateItem eSpecStateItem;
        private ESpecStateId eSpecStateId;

        private boolean minor = false;

        public ESpecLogItemAdder(long id) {
            this.id = id;
        }

        ESpecLogItemAdder setText(@Nonnull String text) {
            this.text = text;
            return this;
        }

        ESpecLogItemAdder setStatus(@Nonnull TaskStatusIndicator.Status status) {
            this.status = status;
            return this;
        }
        ESpecLogItemAdder setFinalStatus(boolean success) {
            if (success) {
                setStatus(TaskStatusIndicator.Status.SUCCESS);
            } else {
                setStatus(TaskStatusIndicator.Status.FAILED);
            }
            return this;
        }
        ESpecLogItemAdder setStatus(ESpecStepStatus.Status status) {
            switch (status) {
                case RUNNING: { setStatus(TaskStatusIndicator.Status.BUSY); break; }
                case SUCCESS: { setStatus(TaskStatusIndicator.Status.SUCCESS); break; }
                case FAILURE: { setStatus(TaskStatusIndicator.Status.FAILED); break; }
                default: throw new RuntimeException("Unsupported status");
            }
            return this;
        }
        ESpecLogItemAdder setStatus(@Nonnull ESpecStepStatus status) {
            setStatus (status.getStatus());
            return this;
        }

        public ESpecLogItemAdder setESpecStateId(ExperimentState experimentState, String especStepId) {
            this.eSpecStateId = new ESpecStateId(experimentState, especStepId);
            return this;
        }
        public ESpecLogItemAdder setExperimentState(ExperimentState experimentState) {
            this.eSpecStateId = new ESpecStateId(experimentState, null);
            return this;
        }
        public ESpecLogItemAdder setESpecStepId(String especStepId) {
            this.eSpecStateId = new ESpecStateId(ExperimentState.SETUP_SOFTWARE, especStepId);;
            return this;
        }
        public ESpecLogItemAdder setESpecStateId(ESpecStateId eSpecStateId) {
            this.eSpecStateId = eSpecStateId;
            return this;
        }

        ESpecLogItemAdder setESpecStateItem(ESpecStateItem eSpecStateItem) {
            this.eSpecStateItem = eSpecStateItem;
            return this;
        }

        //set base state to whatever the current base state is
        ESpecLogItemAdder setCurrentBaseState() {
            this.eSpecStateId = new ESpecStateId(experiment.getExperimentState(), null);
            return this;
        }

        //set base state to whatever the current base state is
        ESpecLogItemAdder setIsMinor() {
            this.minor = true;
            return this;
        }
        ESpecLogItemAdder setIsMinor(boolean newMinor) {
            this.minor = newMinor;
            return this;
        }

        public ESpecLogItemAdder setEspecStep(ESpecStep especStep) {
            this.especStep = especStep;
            return this;
        }

        public ESpecLogItemAdder setLimitedLiveLog(LimitedLiveLog limitedLiveLog) {
            this.limitedLiveLog = limitedLiveLog;
            return this;
        }

        public ESpecLogItemAdder setUploadProgressTracker(UploadProgressTracker uploadProgressTracker) {
            this.uploadProgressTracker = uploadProgressTracker;
            return this;
        }

        public void addOrUpdate() {
            if (text == null) throw new IllegalStateException("text should have been set");
            if (status == null) throw new IllegalStateException("status should have been set");
            if (eSpecStateId == null && eSpecStateItem == null) throw new IllegalStateException("eSpecStateId or eSpecStateItem should have been set");
            ESpecLogItem res = new ESpecLogItem(id, text, status, especStep, uploadProgressTracker, limitedLiveLog);

            if (eSpecStateId != null) {
                eSpecStateItem = getState(eSpecStateId, ESpecStateItem.ESpecStateStatus.RUNNING);
                //noinspection ConstantConditions
                if (eSpecStateItem == null) {
                    throw new IllegalStateException("Should never happen");
                }
            }
            if (eSpecStateItem != null) {
                res.setESpecStateItem(eSpecStateItem);
            }

            Platform.runLater(() -> {
                addOrUpdateESpecLogItem(res);
            });
        }
    }

    private void addOrUpdateESpecLogItem(@Nonnull ESpecLogItem eSpecLogItem) {
        //TODO: use something more efficient than contains
        ESpecLogItem existing = eSpecLogItemsById.get(eSpecLogItem.getId());
        if (existing != null) {
            existing.update(eSpecLogItem);
        } else {
            eSpecLogItemsById.put(eSpecLogItem.getId(), eSpecLogItem);
            eSpecLogItems.add(eSpecLogItem);
            if (eSpecLogItem.getESpecStateItem() != null) {
                eSpecLogItem.getESpecStateItem().getLogItems().add(eSpecLogItem);
            }
        }
    }
    private ESpecLogItemAdder addLog() {
        return new ESpecLogItemAdder(logger.getNextLogId());
    }
    private ESpecLogItemAdder addLog(long id) {
        return new ESpecLogItemAdder(id);
    }



    @Override
    public void onPreFileLoad() {

    }

    @Override
    public void onPostFileLoad(@Nonnull FileSource fileSource,
                               boolean isFast, long byteSize,
                               @Nonnull ESpecStepStatus status) {
        String fileText = fileSource.getBasename();
        addLog().setText("File load \""+fileText+"\"").setStatus(status)
                .setCurrentBaseState()
                .setIsMinor(!status.isFailure())
                .addOrUpdate();
    }

    @Override
    public void onPostFileLoadAll(boolean success) {
        addLog().setText("All files loaded").setFinalStatus(success)
                .setCurrentBaseState()
                .addOrUpdate();
    }

    @Override
    public void onPreRSpec() {

    }

    @Override
    public void onRequestRSpecKnown(@Nonnull RspecSpec rspecSpec, @Nonnull String requestRspec) {
        addLog().setText("Request RSpec Available").setFinalStatus(true)
                .setCurrentBaseState()
                .setEspecStep(rspecSpec)
                .addOrUpdate();
    }

    @Override
    public void onPostRSpec(@Nonnull RspecSpec rspec, @Nonnull ESpecStepStatus status) {
        if (status.isFailure())
            addLog().setText("RSpec Provision").setStatus(status)
                    .setCurrentBaseState()
                    .setEspecStep(rspec)
                    .addOrUpdate();
    }

    @Override
    public void onPostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds) {
    }

    @Override
    public void onPreDir() {
        setPreviousRunningToSuccess(ExperimentState.SETUP_SOFTWARE, "Dir");
        setStatus(ExperimentState.SETUP_SOFTWARE, "Dir", ESpecStateItem.ESpecStateStatus.RUNNING);
    }

    @Override
    public void onPreDirNode(@Nonnull String nodeClientId) {
    }

    @Override
    public void onPreDir(long logEventId, @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId) {
        addLog(logEventId).setText("Dir \""+fullPath+"\" Creation at "+nodeClientId).setStatus(ESpecStepStatus.Status.RUNNING)
                .setESpecStepId("Dir")
                .setEspecStep(dir)
                .addOrUpdate();
    }

    @Override
    public void onPostDir(long logEventId,
                          @Nonnull DirSpec dir,
                          @Nonnull String fullPath,
                          @Nonnull String nodeClientId,
                          @Nonnull ESpecStepStatus status) {
        addLog(logEventId).setText("Dir \""+fullPath+"\" Creation at "+nodeClientId).setStatus(status)
                .setESpecStepId("Dir")
                .setEspecStep(dir)
                .addOrUpdate();
    }

    @Override
    public void onPostDirNode(@Nonnull String nodeClientId, boolean success) {
        addLog().setText("Finished Dir Creations at "+nodeClientId).setFinalStatus(success)
                .setESpecStepId("Dir")
                .setIsMinor()
                .addOrUpdate();
    }

    @Override
    public void onPostDirAll(boolean success) {
        setStatus(ExperimentState.SETUP_SOFTWARE, "Dir", success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
        addLog().setText("Finished Dir Creations").setFinalStatus(success)
                .setESpecStepId("Dir")
                .addOrUpdate();
    }

    @Override
    public void onPreUpload() {
        setPreviousRunningToSuccess(ExperimentState.SETUP_SOFTWARE, "Upload");
        setStatus(ExperimentState.SETUP_SOFTWARE, "Upload", ESpecStateItem.ESpecStateStatus.RUNNING);
    }

    @Override
    public void onPreUploadNode(@Nonnull String nodeClientId) {

    }

    @Override
    public void onPreUpload(long logEventId, @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId, @Nonnull UploadProgressTracker uploadProgressTracker) {
        String path = uploadSpec.getPath() == null ? fullPath : uploadSpec.getPath();
        addLog(logEventId).setText("Upload \""+path+"\" at "+nodeClientId).setStatus(ESpecStepStatus.Status.RUNNING)
                .setESpecStepId("Upload")
                .setUploadProgressTracker(uploadProgressTracker)
                .setEspecStep(uploadSpec)
                .addOrUpdate();
    }

    @Override
    public void onPostUpload(long logEventId,
                             @Nonnull UploadLikeSpec uploadSpec,
                             @Nonnull String fullPath,
                             @Nonnull String nodeClientId,
                             @Nonnull ESpecStepStatus status) {
        String path = uploadSpec.getPath() == null ? fullPath : uploadSpec.getPath();
        addLog(logEventId).setText("Upload \""+path+"\" at "+nodeClientId).setStatus(status)
                .setESpecStepId("Upload")
                .setEspecStep(uploadSpec)
                .addOrUpdate();
    }

    @Override
    public void onPostUploadNode(@Nonnull String nodeClientId, boolean success) {
        addLog().setText("Finished Uploads at "+nodeClientId).setFinalStatus(success)
                .setESpecStepId("Upload")
                .setIsMinor()
                .addOrUpdate();
    }

    @Override
    public void onPostUploadAll(boolean success) {
        setStatus(ExperimentState.SETUP_SOFTWARE, "Upload",
                success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
        addLog().setText("Finished Uploads").setFinalStatus(success)
                .setESpecStepId("Upload")
                .addOrUpdate();
    }

    @Override
    public void onPreExecute() {
        setPreviousRunningToSuccess(ExperimentState.SETUP_SOFTWARE, "Execute");
        setStatus(ExperimentState.SETUP_SOFTWARE, "Execute", ESpecStateItem.ESpecStateStatus.RUNNING);
    }

    @Override
    public void onPreExecute(long logEventId,
                             @Nonnull ExecuteSpec executeSpec,
                             @Nonnull String fullExecutablePath,
                             @Nullable String fullLogPath,
                             @Nonnull String nodeClientId,
                             @Nonnull LimitedLiveLog currentLog) {
        String path = executeSpec.getPath() == null ? fullExecutablePath : executeSpec.getPath();
        addLog(logEventId).setText("Executing \""+path+"\" at "+nodeClientId).setStatus(ESpecStepStatus.Status.RUNNING)
                .setESpecStepId("Execute")
                .setLimitedLiveLog(currentLog)
                .setEspecStep(executeSpec)
                .addOrUpdate();
    }

    @Override
    public void onPostExecute(long logEventId,
                              @Nonnull ExecuteSpec executeSpec,
                              @Nonnull String fullExecutablePath,
                              @Nullable String fullLogPath,
                              @Nonnull String nodeClientId,
                              int exitStatus,
                              @Nonnull ESpecStepStatus status) {
        String path = executeSpec.getPath() == null ? fullExecutablePath : executeSpec.getPath();
        addLog(logEventId).setText("Execute \""+path+"\" at "+nodeClientId).setStatus(status)
                .setESpecStepId("Execute")
                .setEspecStep(executeSpec)
                .addOrUpdate();
    }

    @Override
    public void onPostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success) {

    }

    @Override
    public void onPostExecuteAll(boolean success) {
        setStatus(ExperimentState.SETUP_SOFTWARE, "Execute",
                success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
        addLog().setText("Finished Executes").setFinalStatus(success)
                .setESpecStepId("Execute")
                .addOrUpdate();
    }

    @Override
    public void onPreAnsible() {
        setPreviousRunningToSuccess(ExperimentState.SETUP_SOFTWARE, "Ansible");
        setStatus(ExperimentState.SETUP_SOFTWARE, "Ansible", ESpecStateItem.ESpecStateStatus.RUNNING);
    }

    @Override
    public void onPreAnsiblePlaybook(long logEventId,
                                     @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                     @Nonnull String fullPlaybookPath,
                                     @Nullable String fullLogPath,
                                     @Nonnull String nodeClientId,
                                     @Nonnull LimitedLiveLog currentLog) {
        addLog(logEventId).setText("Ansible Playbook \""+fullPlaybookPath+"\" at "+nodeClientId).setStatus(ESpecStepStatus.Status.RUNNING)
                .setESpecStepId("Ansible")
                .setLimitedLiveLog(currentLog)
                .setEspecStep(ansiblePlaybookSpec)
                .addOrUpdate();
    }

    @Override
    public void onPostAnsiblePlaybook(long logEventId,
                                      @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                      @Nonnull String fullPlaybookPath,
                                      @Nullable String fullLogPath,
                                      @Nonnull String nodeClientId,
                                      @Nonnull ESpecStepStatus status) {
        addLog(logEventId).setText("Ansible Playbook \""+fullPlaybookPath+"\" at "+nodeClientId).setStatus(status)
                .setESpecStepId("Ansible")
                .setEspecStep(ansiblePlaybookSpec)
                .addOrUpdate();
    }

    @Override
    public void onPostAnsibleAll(boolean success) {
        setStatus(ExperimentState.SETUP_SOFTWARE, "Ansible",
                success ? ESpecStateItem.ESpecStateStatus.SUCCESS : ESpecStateItem.ESpecStateStatus.FAILURE);
        addLog().setText("Ansible Playbooks").setFinalStatus(success)
                .setESpecStepId("Ansible")
                .addOrUpdate();
    }
}
